@@ -12,7 +12,7 @@ use Encode;
 sub printheader ($) {
 	my $session=shift;
 	
-	if ($config{sslcookie}) {
+	if (($ENV{HTTPS} && lc $ENV{HTTPS} ne "off") || $config{sslcookie}) {
 		print $session->header(-charset => 'utf-8',
 			-cookie => $session->cookie(-httponly => 1, -secure => 1));
 	}
@@ -46,13 +46,68 @@ sub showform ($$$$;@) {
 	my $cgi=shift;
 
 	printheader($session);
-	print misctemplate($form->title, $form->render(submit => $buttons), @_);
+	print cgitemplate($cgi, $form->title,
+		$form->render(submit => $buttons), @_);
+}
+
+sub cgitemplate ($$$;@) {
+	my $cgi=shift;
+	my $title=shift;
+	my $content=shift;
+	my %params=@_;
+	
+	my $template=template("page.tmpl");
+
+	my $topurl = $config{url};
+	if (defined $cgi && ! $config{w3mmode} && ! $config{reverse_proxy}) {
+		$topurl = $cgi->url;
+	}
+
+	my $page="";
+	if (exists $params{page}) {
+		$page=delete $params{page};
+		$params{forcebaseurl}=urlto($page);
+		if (! $config{html5}) {
+			$params{forcebaseurl}=urlabs($params{forcebaseurl}, $topurl);
+		}
+	}
+	run_hooks(pagetemplate => sub {
+		shift->(
+			page => $page,
+			destpage => $page,
+			template => $template,
+		);
+	});
+	templateactions($template, "");
+
+	my $baseurl = baseurl();
+	if (! $config{html5}) {
+		$baseurl = urlabs($baseurl, $topurl),
+	}
+
+	$template->param(
+		dynamic => 1,
+		title => $title,
+		wikiname => $config{wikiname},
+		content => $content,
+		baseurl => $baseurl,
+		html5 => $config{html5},
+		%params,
+	);
+	
+	return $template->output;
 }
 
 sub redirect ($$) {
 	my $q=shift;
 	eval q{use URI};
-	my $url=URI->new(shift);
+
+	my $topurl;
+	if (defined $q && ! $config{w3mmode} && ! $config{reverse_proxy}) {
+		$topurl = $q->url;
+	}
+
+	my $url=URI->new(urlabs(shift, $topurl));
 	if (! $config{w3mmode}) {
 		print $q->redirect($url);
 	}
@@ -67,16 +122,29 @@ sub decode_cgi_utf8 ($) {
 	if ($] < 5.01) {
 		my $cgi = shift;
 		foreach my $f ($cgi->param) {
-			$cgi->param($f, map { decode_utf8 $_ } $cgi->param($f));
+			$cgi->param($f, map { decode_utf8 $_ }
+				@{$cgi->param_fetch($f)});
 		}
 	}
 }
 
+sub safe_decode_utf8 ($) {
+    my $octets = shift;
+    # call decode_utf8 on >= 5.20 only if it's not already decoded,
+    # otherwise it balks, on < 5.20, always call it
+    if ($] < 5.02 || !Encode::is_utf8($octets)) {
+        return decode_utf8($octets);
+    }
+    else {
+        return $octets;
+    }
+}
+
 sub decode_form_utf8 ($) {
 	if ($] >= 5.01) {
 		my $form = shift;
 		foreach my $f ($form->field) {
-			my @value=map { decode_utf8($_) } $form->field($f);
+			my @value=map { safe_decode_utf8($_) } $form->field($f);
 			$form->field(name  => $f,
 			             value => \@value,
 		                     force => 1,
@@ -93,7 +161,7 @@ sub needsignin ($$) {
 
 	if (! defined $session->param("name") ||
 	    ! userinfo_get($session->param("name"), "regdate")) {
-		$session->param(postsignin => $ENV{QUERY_STRING});
+		$session->param(postsignin => $q->query_string);
 		cgi_signin($q, $session);
 		cgi_savesession($session);
 		exit;
@@ -116,7 +184,7 @@ sub cgi_signin ($$;$) {
 		required => 'NONE',
 		javascript => 0,
 		params => $q,
-		action => $config{cgiurl},
+		action => cgiurl(),
 		header => 0,
 		template => {type => 'div'},
 		stylesheet => 1,
@@ -198,7 +266,7 @@ sub cgi_prefs ($$) {
 		required => 'NONE',
 		javascript => 0,
 		params => $q,
-		action => $config{cgiurl},
+		action => cgiurl(),
 		template => {type => 'div'},
 		stylesheet => 1,
 		fieldsets => [
@@ -231,11 +299,11 @@ sub cgi_prefs ($$) {
 	
 	if ($form->submitted eq 'Logout') {
 		$session->delete();
-		redirect($q, $config{url});
+		redirect($q, baseurl(undef));
 		return;
 	}
 	elsif ($form->submitted eq 'Cancel') {
-		redirect($q, $config{url});
+		redirect($q, baseurl(undef));
 		return;
 	}
 	elsif ($form->submitted eq 'Save Preferences' && $form->validate) {
@@ -313,7 +381,8 @@ sub cgi_getsession ($) {
 			{ FileName => "$config{wikistatedir}/sessions.db" })
 	};
 	if (! $session || $@) {
-		error($@." ".CGI::Session->errstr());
+		my $error = $@;
+		error($error." ".CGI::Session->errstr());
 	}
 	
 	umask($oldmask);
@@ -391,7 +460,7 @@ sub cgi (;$$) {
 			# userinfo db.
 			if (! userinfo_get($session->param("name"), "regdate")) {
 				userinfo_setall($session->param("name"), {
-					email => "",
+					email => defined $session->param("email") ? $session->param("email") : "",
 					password => "",
 					regdate => time,
 				}) || error("failed adding user");
@@ -423,7 +492,7 @@ sub cgierror ($) {
 	my $message=shift;
 
 	print "Content-type: text/html\n\n";
-	print misctemplate(gettext("Error"),
+	print cgitemplate(undef, gettext("Error"),
 		"<p class=\"error\">".gettext("Error").": $message</p>");
 	die $@;
 }
