@@ -8,7 +8,6 @@ use IkiWiki 3.00;
 use HTML::Parser;
 use HTML::Tagset;
 use HTML::Entities;
-use URI;
 use open qw{:utf8 :std};
 
 my %feeds;
@@ -17,7 +16,8 @@ my %guids;
 sub import {
 	hook(type => "getopt", id => "aggregate", call => \&getopt);
 	hook(type => "getsetup", id => "aggregate", call => \&getsetup);
-	hook(type => "checkconfig", id => "aggregate", call => \&checkconfig);
+	hook(type => "checkconfig", id => "aggregate", call => \&checkconfig,
+		last => 1);
 	hook(type => "needsbuild", id => "aggregate", call => \&needsbuild);
 	hook(type => "preprocess", id => "aggregate", call => \&preprocess);
         hook(type => "delete", id => "aggregate", call => \&delete);
@@ -65,6 +65,8 @@ sub checkconfig () {
 		$config{aggregateinternal}=1;
 	}
 
+	# This is done here rather than in a refresh hook because it
+	# needs to run before the wiki is locked.
 	if ($config{aggregate} && ! ($config{post_commit} && 
 	                             IkiWiki::commit_hook_enabled())) {
 		launchaggregation();
@@ -102,8 +104,7 @@ sub launchaggregation () {
 	my @feeds=needsaggregate();
 	return unless @feeds;
 	if (! lockaggregate()) {
-		debug("an aggregation process is already running");
-		return;
+		error("an aggregation process is already running");
 	}
 	# force a later rebuild of source pages
 	$IkiWiki::forcerebuild{$_->{sourcepage}}=1
@@ -163,10 +164,14 @@ sub migrate_to_internal {
 		
 		$config{aggregateinternal} = 0;
 		my $oldname = "$config{srcdir}/".htmlfn($data->{page});
+		if (! -e $oldname) {
+			$oldname = $IkiWiki::Plugin::transient::transientdir."/".htmlfn($data->{page});
+		}
+
 		my $oldoutput = $config{destdir}."/".IkiWiki::htmlpage($data->{page});
 		
 		$config{aggregateinternal} = 1;
-		my $newname = "$config{srcdir}/".htmlfn($data->{page});
+		my $newname = $IkiWiki::Plugin::transient::transientdir."/".htmlfn($data->{page});
 		
 		debug "moving $oldname -> $newname";
 		if (-e $newname) {
@@ -186,7 +191,7 @@ sub migrate_to_internal {
 		if (-e $oldoutput) {
 			require IkiWiki::Render;
 			debug("removing output file $oldoutput");
-			IkiWiki::prune($oldoutput);
+			IkiWiki::prune($oldoutput, $config{destdir});
 		}
 	}
 	
@@ -210,6 +215,8 @@ sub needsbuild (@) {
 			markunseen($feed->{sourcepage});
 		}
 	}
+
+	return $needsbuild;
 }
 
 sub preprocess (@) {
@@ -384,13 +391,16 @@ sub garbage_collect () {
 	foreach my $guid (values %guids) {
 		# any guid whose feed is gone should be removed
 		if (! exists $feeds{$guid->{feed}}) {
-			unlink "$config{srcdir}/".htmlfn($guid->{page})
-				if exists $guid->{page};
+			if (exists $guid->{page}) {
+				unlink $IkiWiki::Plugin::transient::transientdir."/".htmlfn($guid->{page})
+					|| unlink "$config{srcdir}/".htmlfn($guid->{page});
+			}
 			delete $guids{$guid->{guid}};
 		}
 		# handle expired guids
 		elsif ($guid->{expired} && exists $guid->{page}) {
 			unlink "$config{srcdir}/".htmlfn($guid->{page});
+			unlink $IkiWiki::Plugin::transient::transientdir."/".htmlfn($guid->{page});
 			delete $guid->{page};
 			delete $guid->{md5};
 		}
@@ -478,6 +488,7 @@ sub needsaggregate () {
 }
 
 sub aggregate (@) {
+	eval q{use Net::INET6Glue::INET_is_INET6}; # may not be available
 	eval q{use XML::Feed};
 	error($@) if $@;
 	eval q{use URI::Fetch};
@@ -502,7 +513,8 @@ sub aggregate (@) {
 			}
 			$feed->{feedurl}=pop @urls;
 		}
-		my $res=URI::Fetch->fetch($feed->{feedurl});
+		my $ua=useragent();
+		my $res=URI::Fetch->fetch($feed->{feedurl}, UserAgent=>$ua);
 		if (! $res) {
 			$feed->{message}=URI::Fetch->errstr;
 			$feed->{error}=1;
@@ -541,7 +553,9 @@ sub aggregate (@) {
 			};
 		}
 		if ($@) {
-			$feed->{message}=gettext("feed crashed XML::Feed!")." ($@)";
+			# gettext can clobber $@
+			my $error = $@;
+			$feed->{message}=gettext("feed crashed XML::Feed!")." ($error)";
 			$feed->{error}=1;
 			debug($feed->{message});
 			next;
@@ -569,6 +583,7 @@ sub aggregate (@) {
 				feed => $feed,
 				copyright => $f->copyright,
 				title => defined $entry->title ? decode_entities($entry->title) : "untitled",
+				author => defined $entry->author ? decode_entities($entry->author) : "",
 				link => $entry->link,
 				content => (defined $c && defined $c->body) ? $c->body : "",
 				guid => defined $entry->id ? $entry->id : time."_".$feed->{name},
@@ -589,6 +604,7 @@ sub add_page (@) {
 		# updating an existing post
 		$guid=$guids{$params{guid}};
 		return if $guid->{expired};
+		write_page($feed, $guid, $mtime, \%params);
 	}
 	else {
 		# new post
@@ -610,25 +626,39 @@ sub add_page (@) {
 		}
 		my $c="";
 		while (exists $IkiWiki::pagecase{lc $page.$c} ||
+		       -e $IkiWiki::Plugin::transient::transientdir."/".htmlfn($page.$c) ||
 		       -e "$config{srcdir}/".htmlfn($page.$c)) {
 			$c++
 		}
+		$page=$page.$c;
 
-		# Make sure that the file name isn't too long. 
-		# NB: This doesn't check for path length limits.
-		my $max=POSIX::pathconf($config{srcdir}, &POSIX::_PC_NAME_MAX);
-		if (defined $max && length(htmlfn($page)) >= $max) {
+		$guid->{page}=$page;
+		eval { write_page($feed, $guid, $mtime, \%params) };
+		if ($@) {
+			# assume failure was due to a too long filename
 			$c="";
 			$page=$feed->{dir}."/item";
 			while (exists $IkiWiki::pagecase{lc $page.$c} ||
-			       -e "$config{srcdir}/".htmlfn($page.$c)) {
+			      -e $IkiWiki::Plugin::transient::transientdir."/".htmlfn($page.$c) ||
+			      -e "$config{srcdir}/".htmlfn($page.$c)) {
 				$c++
 			}
+			$page=$page.$c;
+
+			$guid->{page}=$page;
+			write_page($feed, $guid, $mtime, \%params);
 		}
 
-		$guid->{page}=$page;
 		debug(sprintf(gettext("creating new page %s"), $page));
 	}
+}
+
+sub write_page ($$$$$) {
+	my $feed=shift;
+	my $guid=shift;
+	my $mtime=shift;
+	my %params=%{shift()};
+
 	$guid->{feed}=$feed->{name};
 	
 	# To write or not to write? Need to avoid writing unchanged pages
@@ -642,27 +672,40 @@ sub add_page (@) {
 	$guid->{md5}=$digest;
 
 	# Create the page.
-	my $template=template($feed->{template}, blind_cache => 1);
+	my $template;
+	eval {
+		$template=template($feed->{template}, blind_cache => 1);
+	};
+	if ($@) {
+		# gettext can clobber $@
+		my $error = $@;
+		print STDERR gettext("failed to process template:")." $error";
+		return;
+	}
 	$template->param(title => $params{title})
 		if defined $params{title} && length($params{title});
+	$template->param(author => $params{author})
+		if defined $params{author} && length($params{author}
+			&& $params{author} ne $feed->{name});
 	$template->param(content => wikiescape(htmlabs($params{content},
 		defined $params{base} ? $params{base} : $feed->{feedurl})));
 	$template->param(name => $feed->{name});
 	$template->param(url => $feed->{url});
 	$template->param(copyright => $params{copyright})
 		if defined $params{copyright} && length $params{copyright};
-	$template->param(permalink => urlabs($params{link}, $feed->{feedurl}))
+	$template->param(permalink => IkiWiki::urlabs($params{link}, $feed->{feedurl}))
 		if defined $params{link};
 	if (ref $feed->{tags}) {
 		$template->param(tags => [map { tag => $_ }, @{$feed->{tags}}]);
 	}
-	writefile(htmlfn($guid->{page}), $config{srcdir},
-		$template->output);
+	writefile(htmlfn($guid->{page}),
+		$IkiWiki::Plugin::transient::transientdir, $template->output);
 
 	if (defined $mtime && $mtime <= time) {
 		# Set the mtime, this lets the build process get the right
 		# creation time on record for the new page.
-		utime $mtime, $mtime, "$config{srcdir}/".htmlfn($guid->{page});
+		utime $mtime, $mtime,
+			$IkiWiki::Plugin::transient::transientdir."/".htmlfn($guid->{page});
 		# Store it in pagectime for expiry code to use also.
 		$IkiWiki::pagectime{$guid->{page}}=$mtime
 			unless exists $IkiWiki::pagectime{$guid->{page}};
@@ -679,13 +722,6 @@ sub wikiescape ($) {
 	return encode_entities(shift, '\[\]');
 }
 
-sub urlabs ($$) {
-	my $url=shift;
-	my $urlbase=shift;
-
-	URI->new_abs($url, $urlbase)->as_string;
-}
-
 sub htmlabs ($$) {
 	# Convert links in html from relative to absolute.
 	# Note that this is a heuristic, which is not specified by the rss
@@ -711,7 +747,7 @@ sub htmlabs ($$) {
 				next unless $v_offset; # 0 v_offset means no value
 				my $v = substr($text, $v_offset, $v_len);
 				$v =~ s/^([\'\"])(.*)\1$/$2/;
-				my $new_v=urlabs($v, $urlbase);
+				my $new_v=IkiWiki::urlabs($v, $urlbase);
 				$new_v =~ s/\"/&quot;/g; # since we quote with ""
 				substr($text, $v_offset, $v_len) = qq("$new_v");
 			}
